#ifndef M2IF_VECTOR_H
#define M2IF_VECTOR_H

#include <cmath>
#include <ostream>

namespace m2if {

template <class T>
class Vector
{
 private:
  int size_;
  T *data_;

 public: // structors
  Vector(int size);
  Vector(const Vector & v);
  ~Vector();

 public: // setters
  Vector & operator=(const Vector & v);
  T & operator[](int idx);

 public: // getters
  int get_size() const { return size_; }
  int get_size() { return size_; }
  T operator[](int idx) const;

 public: // algebra
  template <class U> friend Vector<U> operator+(const Vector<U> & v1, const Vector<U> & v2);
  template <class U> friend Vector<U> operator-(const Vector<U> & v1, const Vector<U> & v2);
  template <class U> friend U operator*(const Vector<U> & v1, const Vector<U> & v2);
  T norm() { return sqrt(*this * *this); }

 public: // stream
  template <class U> friend std::ostream & operator<<(std::ostream & st, const Vector<U> & v);
};

// structors

template <class T>
Vector<T>::Vector(int size)
{
  if (size <= 0)
    throw "wrong size";

  size_ = size;
  data_ = new T[size_];
  if (!data_)
    throw "fail to allocate memory";
}

template <class T>
Vector<T>::Vector(const Vector<T> & v)
{
  size_ = v.size_;
  data_ = new T[size_];
  if (!data_)
    throw "fail to allocate memory";
  for (int i = 0; i < size_; i++)
    data_[i] = v.data_[i];
}

template <class T>
Vector<T>::~Vector()
{
  delete[] data_;
}

// setters

template <class T>
Vector<T> & Vector<T>::operator=(const Vector<T> & v)
{
  if (&v != this)
    {
      Vector<T> tmp = v;
      T *d = tmp.data_;
      tmp.data_ = data_;
      data_ = d;
    }

  return *this;
}

template <class T>
T & Vector<T>::operator[](int idx)
{
  if ((idx < 0) || (idx >= size_))
    throw "wrong index";

  return data_[idx];
}

// getters

template <class T>
T Vector<T>::operator[](int idx) const
{
  if ((idx < 0) || (idx >= size_))
    throw "wrong index";

  return data_[idx];
}

template<class T> Vector<T> operator+(const Vector<T> & v1, const Vector<T> & v2)
{
  if (v1.get_size() != v2.get_size())
    throw "wrong size";

  Vector<T> v(v1.get_size());

  for (int i = 0; i < v.get_size(); i++)
    v[i]= v1[i] + v2[i];

  return v;
}

template<class T> Vector<T> operator-(const Vector<T> & v1, const Vector<T> & v2)
{
  if (v1.get_size() != v2.get_size())
    throw "wrong size";

  Vector<T> v(v1.get_size());

  for (int i = 0; i < v.get_size(); i++)
    v[i]= v1[i] - v2[i];

  return v;
}

template<class T> T operator*(const Vector<T> & v1, const Vector<T> & v2)
{
  if (v1.get_size() != v2.get_size())
    throw "wrong size";

  T res = 0;
  for (int i = 0; i < v1.get_size(); ++i)
    res += v1[i] * v2[i];

  return res;
}

template<class T> std::ostream & operator<<(std::ostream & st, const Vector<T> & v)
{
  st << "(";
  for (int i = 0; i < v.get_size(); i++)
    {
      st << v[i];
      if (i != (v.get_size() - 1))
	st << ",";
    }
  st << ")";

  return st;
}

} // namespace m2if

#endif
